/* Complex object shadows */
/* An OpenGL demo by A.P.Gardner */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>

/* definitions */

#define YELLOWMAT 1
#define REDMAT 2
#define BLUEMAT 3
#define GREYMAT 4

/* Function to initialise params */

void init(void)
{
	GLfloat yellow_diffuse[] = {0.7, 0.7, 0.0, 1.0};
	GLfloat yellow_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat red_diffuse[] = {1.0, 0.0, 0.0, 1.0};
	GLfloat red_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat blue_diffuse[] = {0.1, 0.1, 0.7, 1.0};
	GLfloat blue_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat grey_diffuse[] = {0.2, 0.2, 0.2, 1.0};
	GLfloat grey_specular[] = {0.0, 0.0, 0.0, 0.0};
	GLfloat ambient_light[] = {0.5, 0.5, 0.5, 1.0};
	GLfloat position_one[] = {0.0, 0.9, 0.4, 0.0};

	/* create materials for objects in the scene */

	glNewList(YELLOWMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, yellow_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, yellow_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 64.0);
	glColor3fv(yellow_diffuse);
	glEndList();

	glNewList(REDMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, red_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, red_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 64.0);
	glColor3fv(red_diffuse);
	glEndList();

	glNewList(BLUEMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, blue_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 45.0);
	glColor3fv(blue_diffuse);
	glEndList();

	glNewList(GREYMAT, GL_COMPILE);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, grey_diffuse);
	glMaterialfv(GL_FRONT, GL_SPECULAR, grey_specular);
	glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
	glColor3fv(grey_diffuse);
	glEndList();

	/* set lighting params */

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);
	glLightfv(GL_LIGHT0, GL_POSITION, position_one);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

	/* enable depth buffering */

	glEnable(GL_DEPTH_TEST);
}

/* Function to draw object and its shadow */

void draw_object()
{
	/* draw object */

	glCallList(BLUEMAT);
	glutSolidCone(0.5, 1.0, 30, 30);

	/* set new render state */

	glDisable(GL_LIGHTING);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_LEQUAL);

	/* translate to shadow position and collapse scale in y axis */

	glTranslatef(0.0, -0.95, 0.0);
	glScalef(1.0, 0.0, 1.0);

	/* render collapsed object in dark colour */

	glCallList(GREYMAT);
	glutSolidCone(0.5, 1.0, 30, 30);

	/* reset rendering state to default */

	glDepthFunc(GL_LESS);
	glDepthMask(GL_TRUE);
	glEnable(GL_LIGHTING);
}

/* Function to draw display */

void display(void)
{
	/* clear all buffers */

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* draw floor & wall panel */

	glStencilFunc(GL_ALWAYS, 0, 0);
	glCallList(YELLOWMAT);
	glBegin(GL_QUADS);
		glNormal3f(0.0, 1.0, 0.0);
		glVertex3f(3.0, -1.0, 4.5);
		glVertex3f(3.0, -1.0, -5.0);
		glVertex3f(-3.0, -1.0, -5.0);
		glVertex3f(-3.0, -1.0, 4.5);
	glEnd();
	glCallList(REDMAT);
	glBegin(GL_QUADS);
		glNormal3f(0.0, 0.0, 1.0);
		glVertex3f(-3.0, 1.0, 1.0);
		glVertex3f(-1.0, 1.0, 1.0);
		glVertex3f(-1.0, -1.0, 1.0);
		glVertex3f(-3.0, -1.0, 1.0);
	glEnd();

	/* draw 1st object & shadow */

	glPushMatrix();
	glTranslatef(-1.0, 0.0, 0.0);
	draw_object();
	glPopMatrix();

	/* draw 2nd object & shadow */

	glPushMatrix();
	glTranslatef(1.0, 0.0, 0.0);
	draw_object();
	glPopMatrix();

	/* process all buffered OpenGL commands */

	glFlush();
}

/* Function to handle window size changes */

void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (GLfloat)w/(GLfloat)h, 3.0, 7.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -5.0);
}

/* Main function */

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_SINGLE | GLUT_RGB);
	glutInitWindowSize(500, 500);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("OpenGL Demo - Complex Shadows");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMainLoop();

	return 0;
}